/* allocate from a static buffer until it is exhausted, subsequently allocate
   from the heap */

#include <malloc.h>

#undef malloc
#undef calloc
#undef realloc
#undef free

#ifndef BE_ALLOCATION_SIZE
#define BE_ALLOCATION_SIZE (1 << 20)
#endif

static char hunk_o_mem[BE_ALLOCATION_SIZE];
static int start = 0;

int use_heap = 0; /* we check these in bemap.c */
int allocated = 0;
int freed = 0;

/* this tells us if a pointer points into our static buffer or not */
#define STATIC_ADDRESS(x) ((x >= hunk_o_mem) && (x < hunk_o_mem+BE_ALLOCATION_SIZE))

/* Allocate num_bytes of memory */
void* be_malloc (size_t num_bytes)
{
    char* result;
    int chunk_size;

    /* round num_bytes up to 8 byte chunk */
    chunk_size = num_bytes + num_bytes % 8;
    allocated += chunk_size + sizeof(size_t);

    /* stop grabbing from the static chunk if it's full */
    if (use_heap) return malloc(num_bytes);

    /* make sure we've got memory enough to satisfy this request */
    if (num_bytes > BE_ALLOCATION_SIZE-start) {
        use_heap = 1;
        return malloc(num_bytes);
    }

    result = hunk_o_mem + start + sizeof(size_t);
    start += chunk_size + sizeof(size_t);

    /* yes this looks funky, but i know what I'm doing */
    ((size_t*)result)[-1] = num_bytes;

    return result;
}

/* Re-allocate the previously allocated block in old_ptr, making the new block
   new_size bytes long */
void* be_realloc (void* old_ptr, size_t new_size)
{
    if (!use_heap || STATIC_ADDRESS(old_ptr)) {
        char* new_data = be_malloc(new_size);
        memcpy(new_data, old_ptr, ((size_t*)old_ptr)[-1]);
        return new_data;

    } else {
        return realloc(old_ptr, new_size);
    }
}

/* Allocate num_elements elements of size bytes each, all initialized to 0 */
void* be_calloc (size_t num_elements, size_t size)
{
    char* new_data;

    if (use_heap) return calloc(num_elements, size);

    new_data = be_malloc(num_elements*size);
    memset(new_data, 0, num_elements*size);

    return new_data;
}

/* Free a block allocated by `malloc', `realloc' or `calloc' */
void be_free (void* ptr)
{
    /* pass this on to the real free if we didn't malloc it */
    if (!STATIC_ADDRESS(ptr)) free(ptr);

    /* otherwise we do nothing. no free's here kids. */
    else freed += ((size_t*)ptr)[-1] + sizeof(size_t);
}
